# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Collection of tools for collection of statistical data from a model.

Objects of interest are those that we see people working on most. We
think that counting those may help us with model complexity evaluation -
for example identify if a model is big or small or see where the
modeling focus is (problem space / solution space / balanced)
"""

import capellambse

COMMON_OBJECTS = [
    "Requirement",
    "Class",
    "StateMachine",
    "PhysicalLink",
    "FunctionalExchange",
    "ComponentExchange",
    "ExchangeItem",
]

OBJECTS_OF_INTEREST = [
    [  # Operational Analysis
        "OperationalCapability",
        "OperationalActivity",
        "Entity",
    ]
    + COMMON_OBJECTS,
    ["SystemFunction", "SystemComponent"] + COMMON_OBJECTS,  # System Analysis
    [  # Logical Architecture
        "LogicalFunction",
        "LogicalComponent",
    ]
    + COMMON_OBJECTS,
    [  # Physical Architecture
        "PhysicalFunction",
        "PhysicalComponent",
    ]
    + COMMON_OBJECTS,
]


def quantify_model_layers(
    model: capellambse.MelodyModel,
) -> tuple[list[int], list[int]]:
    """Count objects of interest and diagrams on model layers.

    Returns
    -------
    list
        The number of interesting objects per model layer
    list
        The number of diagrams per model layer

    Notes
    -----
    The order of numbers in a list corresponds to the order of model
    layers - OA, SA, LA, PA.
    """
    objects = []
    diagrams = []
    for layer, object_types in zip(
        [model.oa, model.sa, model.la, model.pa], OBJECTS_OF_INTEREST
    ):
        layer_objects = len(model.search(*object_types, below=layer))
        objects.append(layer_objects)
        diagrams.append(len(layer.diagrams))
    return objects, diagrams



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

import pathlib

import pytest

from capellambse import helpers


def test_paths_relative_to_root_are_not_changed():
    actual = helpers.normalize_pure_path("foo", base="/")
    expected = pathlib.PurePosixPath("foo")
    assert actual == expected


def test_paths_relative_to_a_given_base_are_made_relative_to_root():
    actual = helpers.normalize_pure_path("bar", base="/foo")
    expected = pathlib.PurePosixPath("foo/bar")
    assert actual == expected


def test_absolute_paths_are_made_relative_to_root():
    actual = helpers.normalize_pure_path("/foo/bar")
    expected = pathlib.PurePosixPath("foo/bar")
    assert actual == expected


def test_base_is_ignored_for_absolute_paths():
    actual = helpers.normalize_pure_path("/foo", base="/bar")
    expected = pathlib.PurePosixPath("foo")
    assert actual == expected


def test_doubledot_removes_the_preceding_component():
    actual = helpers.normalize_pure_path("/foo/../bar")
    expected = pathlib.PurePosixPath("bar")
    assert actual == expected


def test_leading_doubledots_are_ignored():
    actual = helpers.normalize_pure_path("/../../../foo")
    expected = pathlib.PurePosixPath("foo")
    assert actual == expected


@pytest.mark.parametrize(
    "input,expected_output",
    [("&gt;", ">"), ("&lt;", "<"), ("&quot;", '"'), ("&apos;", "'")],
)
def test_flatten_html_unescapes_special_char(
    input: str, expected_output: str
) -> None:
    assert helpers.flatten_html_string(input) == expected_output


def test_flatten_html_strips_images() -> None:
    input = """<img src="data:<>">"""
    assert helpers.flatten_html_string(input) == ""


@pytest.mark.parametrize(
    "input,expected",
    [
        ("First<br>Second", "First\nSecond"),
        ("<p>First</p>Second", "First\nSecond"),
        (
            "<p>First<img>Second<br>Third</p>Fourth",
            "FirstSecond\nThird\nFourth",
        ),
    ],
)
def test_flatten_html_blocks_and_hardbreaks_to_newline(
    input: str, expected: str
) -> None:
    assert helpers.flatten_html_string(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            "<ul><li>item 1</li><li>item 2</li></ul>",
            "             • item 1\n             • item 2",
        ),
        (
            "\n<ul>\n\t<li>item 1</li>\n\t<li>item 2</li></ul>",
            "             • item 1\n             • item 2",
        ),
    ],
)
def test_flatten_html_formats_unordered_lists(
    input: str, expected: str
) -> None:
    assert helpers.flatten_html_string(input) == expected



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Module to populate the diagram cache with native Capella diagrams."""

from __future__ import annotations

import datetime
import json
import logging
import os
import pathlib
import re
import shutil
import subprocess
import tempfile
import typing as t

import lxml
from lxml import etree as ET

import capellambse.model.diagram

E = lxml.builder.ElementMaker()
BAD_FILENAMES = frozenset(
    {"AUX", "CON", "NUL", "PRN"}
    | {f"COM{i}" for i in range(1, 10)}
    | {f"LPT{i}" for i in range(1, 10)}
)
LOGGER = logging.getLogger(__name__)
VALID_IMG_FMTS = frozenset(
    {
        "bmp",
        "gif",
        "jpg",
        "png",
        "svg",
    }
)


class DiagramCache:
    _cli_cmd: list[str] = []
    _cli_type: t.Literal["exe", "docker"] | None = None
    _index: list[dict[str, str | bool]] = []

    def __init__(
        self,
        capella_cli: str,
        image_format: t.Literal["bmp", "gif", "jpg", "png", "svg"],
        create_index: bool,
        force: t.Literal["exe", "docker"] | None,
        capella_project_name: str,
        diagrams_dir: pathlib.Path,
        diagrams: list[capellambse.model.diagram.Diagram],
    ):
        self.capella_cli = capella_cli
        self.image_format = image_format
        self.create_index = create_index
        self.force = force
        self.capella_project_name = capella_project_name
        self.diagrams_dir = diagrams_dir
        self.diagrams = diagrams
        self._json_index_path = self.diagrams_dir / "index.json"
        self._process_config()

    def _find_exe(self, path: str) -> str | None:
        fpath, _ = os.path.split(path)
        if fpath:
            if pathlib.Path(path).is_file() and os.access(path, os.X_OK):
                return path
        else:
            return shutil.which(path)
        return None

    def _process_config(self) -> None:
        if self.image_format.lower() not in VALID_IMG_FMTS:
            raise ValueError(
                f"Invalid image format {self.image_format!r}, must be one of: "
                + ", ".join(sorted(VALID_IMG_FMTS))
            )
        if self.force not in {"docker", "exe", None}:
            raise ValueError(
                "Invalid value for 'force', must be 'docker' or 'exe': "
                f"{self.force}"
            )
        if self.force != "docker":
            # use of Docker is not forced
            if (exe := self._find_exe(self.capella_cli)) is not None:
                # local exe has been identified
                self._cli_cmd = [str(exe)]
                self._cli_type = "exe"
        if not self._cli_cmd and self.force != "exe":
            # no local exe has been identified and Docker is allowed
            if (exe := self._find_exe("docker")) is not None:
                self._cli_cmd = [exe, "run", "--rm"]
                self._cli_type = "docker"
        if not self._cli_cmd:
            if self.force == "exe":
                raise FileNotFoundError(
                    "Cannot find supplied 'capella_cli' executable: "
                    f"{self.capella_cli}"
                )
            elif self.force == "docker":
                raise FileNotFoundError(
                    "Cannot find the executable 'docker' to run the "
                    f"'capella_cli' image: {self.capella_cli}"
                )
            else:
                raise FileNotFoundError(
                    "Cannot run 'capella_cli': Not an executable and "
                    "'docker' was not found to run it as image: "
                    f"{self.capella_cli}"
                )
        self._cli_common_flags = [
            "-nosplash",
            "-consolelog",
            "-application",
            "org.polarsys.capella.core.commandline.core",
            "-appid",
            "org.polarsys.capella.exportRepresentations",
            "-forceimport",
            "-input",
            "/all",
            "-imageFormat",
            f"{self.image_format.upper()}",
            "-outputfolder",
            "/main_model/output",
            "-forceoutputfoldercreation",
        ]

    def _diagram_2_image_file_name(self, fname) -> str:
        # pylint: disable=line-too-long
        r"""Sanitize the filename.

        This function removes all characters that are illegal in file
        names on Windows operating systems, and prefixes reserved names
        with an underscore.

        Note that this also includes the two common directory separators
        (``/`` and ``\``).

        Notes
        -----
        Refer to the `MSDN article about file names`__ for more details.

        __ https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file?redirectedfrom=MSDN#naming-conventions
        """
        # pylint: enable=line-too-long
        fname = fname.rstrip(" .")
        fname = re.sub(
            '[\x00-\x1f<>:"/\\\\|?*]',
            lambda m: "-"[ord(m.group(0)) < 32 :],
            fname,
        )
        if fname.split(".")[0].upper() in BAD_FILENAMES:
            fname = f"_{fname}"
        return fname

    def _rename_diagram_image_files(self) -> None:
        """Rename diagram image files so that the UUID is the base name."""
        existing_diagrams: list[dict[str, str | bool]] = [
            d for d in self._index if d["success"]
        ]
        for diagram_info in existing_diagrams:
            filename: str = str(diagram_info["name"])
            filename = self._diagram_2_image_file_name(
                f"{filename}.{self.image_format}"
            )
            old_path = self.diagrams_dir / filename
            if not old_path.is_file():
                continue
            old_path.rename(
                self.diagrams_dir
                / f"{diagram_info['uuid']}.{self.image_format}"
            )

    def _build_cli_cmd_exe(
        self, tmp_project_dir: pathlib.Path, workspace: pathlib.Path
    ) -> None:
        self._cli_cmd += [
            "-data",
            str(workspace),
            "-import",
            f"{str(tmp_project_dir / 'main_model')}",
        ] + self._cli_common_flags

    def _build_cli_cmd_docker(
        self, tmp_project_dir: pathlib.Path, workspace: pathlib.Path
    ) -> None:
        self._cli_cmd += [
            "-v",
            f"{workspace}:/workspace",
            "-v",
            f"{tmp_project_dir}:/model",
            self.capella_cli,
            "-data",
            "/workspace",
            "-import",
            "/model/main_model",
        ] + self._cli_common_flags

    def _create_index_json(self) -> None:
        """Create a diagram name to UUID index file if not already existing.

        Diagram names in Capella are not unique. When there are multiple
        instances for one diagram name the Capella context menu (.aird node)
        function ``Export representations as images`` works with counting
        suffixes in the image file name.

        The first occurrence of a diagram name leads into a file named
        ``<NAME>.<EXT>``, the second occurrence will be named
        ``<NAME>_1.<EXT>`` etc. whereby ``<EXT>`` might be ``"svg"`` for
        instance.

        Luckily the order of the diagrams that are read by the exporter
        matches the order of diagrams we get from
        :meth:`capellambse.model.MelodyModel.diagrams`.
        """
        diagram_names: dict[str, str] = {}  # uuid -> diag.name
        diagram_name_to_uuid_map: dict[str, str] = {}  # uuid -> diagram name
        diagram_name_counts: dict[
            str, int
        ] = {}  # diag.name -> counts of according diagram name
        for diag in self.diagrams:
            if diag.name in diagram_name_counts:
                diagram_name_counts[diag.name] += 1
            else:
                diagram_name_counts[diag.name] = 1
            diagram_names[diag.uuid] = diag.name
            if diagram_name_counts[diag.name] > 1:
                count_suffix = f"_{diagram_name_counts[diag.name] - 1}"
                diagram_name_to_uuid_map[
                    f"{diag.name}{count_suffix}"
                ] = diag.uuid
            else:
                diagram_name_to_uuid_map[f"{diag.name}"] = diag.uuid
        for diagram_name, uuid in sorted(diagram_name_to_uuid_map.items()):
            filename: str = self._diagram_2_image_file_name(
                f"{diagram_name}.{self.image_format}"
            )
            self._index.append(
                {
                    "name": diagram_name,
                    "uuid": uuid,
                    "success": self.diagrams_dir.joinpath(filename).is_file(),
                }
            )

    def _create_index_html_file(self) -> None:
        """Create an ``index.html`` file linking diagram names to images."""
        filepath = pathlib.Path(self.diagrams_dir / "index.html")
        filepath.unlink(missing_ok=True)
        head = E.head(
            E.style(
                "a:active { text-decoration: none; }"
                " a:hover { text-decoration: none; }"
                " a:link { text-decoration: none; }"
                " a:visited { text-decoration: none; }"
                " li.missing {color: red}"
                " ol {font-family: Courier}"
                " span.small {font-size: 10pt}"
                " span.uuid {color: #dddddd; font-size: 8pt}"
            )
        )
        html = E.html()
        html.append(head)
        now = datetime.datetime.now().strftime("%A, %Y-%m-%d %H:%M:%S")
        title = f"Capella diagram cache for {self.capella_project_name!r}"
        html.append(
            E.h1(
                title,
                E.span(" created: ", now, {"class": "small"}),
            )
        )
        diaglist = E.ol()
        for diagram in self._index:
            label = diagram["name"]
            uuid = E.span(diagram["uuid"], {"class": "uuid"})
            if diagram["success"]:
                href = f"./{diagram['uuid']}.{self.image_format}"
                diaglist.append(E.li(E.a(label, href=href), " ", uuid))
            else:
                diaglist.append(E.li(label, uuid, {"class": "missing"}))
        html.append(diaglist)
        filepath.write_text(ET.tostring(html).decode("ascii"), encoding="utf8")

    def _post_process_svg(self, path: pathlib.Path) -> None:
        """Post process SVG file.

        Parameters
        ----------
        path
            Absolute path to the SVG file of an exported diagram
        """
        tree = ET.parse(path)

        for elm in tree.iter():
            attrib = elm.attrib
            if "font-family" in attrib:
                attrib[
                    "font-family"
                ] = "'Open Sans','Segoe UI',Arial,sans-serif"
            if "stroke-miterlimit" in attrib:
                del attrib["stroke-miterlimit"]

        path.write_bytes(ET.tostring(tree, pretty_print=True))

    def export_diagrams(self, project_dir: pathlib.Path) -> None:
        with tempfile.TemporaryDirectory() as workspace:
            build_cli_cmd = getattr(self, f"_build_cli_cmd_{self._cli_type}")
            build_cli_cmd(project_dir, workspace)
            try:
                subprocess.check_call(self._cli_cmd)
            except subprocess.CalledProcessError as err:
                raise RuntimeError("Failed to update diagram cache") from err
            for p in pathlib.Path(project_dir).glob(
                f"main_model/**/*.{self.image_format}"
            ):
                shutil.copyfile(p, self.diagrams_dir / p.name)
        self._create_index_json()
        self._rename_diagram_image_files()
        if self.image_format == "svg":
            for svg in pathlib.Path(self.diagrams_dir).glob("*.svg"):
                self._post_process_svg(svg)
        if self.create_index:
            self._json_index_path.write_text(json.dumps(self._index, indent=2))
            self._create_index_html_file()



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

import collections.abc as cabc
import errno
import io
import logging
import math
import os
import pathlib
import re
import typing as t
import urllib.parse

import requests

from capellambse import helpers, loader

from . import FileHandler

LOGGER = logging.getLogger(__name__)
RE_LINK_NEXT = re.compile("<(http.*)>; rel=(?P<quote>[\"']?)next(?P=quote)")


class GitlabArtifactsFiles(FileHandler):
    """Download files from Gitlab's artifacts hosting service.

    This file handler is roughly equivalent to an HTTPFileHandler with
    appropriate headers and the following URL::

        https://<path>/api/v4/projects/<project>/jobs/artifacts/<branch>/raw/<subdir>/%s?job_name=<job>

    One important difference is that this file handler will always use
    the latest successful job, regardless of the overall pipeline
    status, while an HTTPFileHandler with the above URL would only
    consider jobs from successful pipelines.

    This file handler uses several of the Gitlab CI/CD `pre-defined
    environment variables`__. Refer to the documentation for their exact
    meaning and behavior during CI runs, and see below for how they are
    used.

    __ https://docs.gitlab.com/ee/ci/variables/predefined_variables.html

    Parameters
    ----------
    path
        The base URL of the Gitlab server. Must start with one of
        ``glart://``, ``glart+http://`` or ``glart+https://``; the
        ``glart:`` prefix uses HTTPS to communicate with the server. May
        also be set to ``glart:`` (without a server name), which uses
        the ``$CI_SERVER_URL`` environment variable to find the
        instance; if that is not set, the public Gitlab instance at
        ``https://gitlab.com`` is used.

        Example: If your project is hosted at
        ``https://gitlab.example.com/my_username/my_cool_project``, use
        ``glart://gitlab.example.com`` as path argument.
    token
        A personal or project access token with ``read_api`` permission
        on the specified project. The following ways are supported for
        passing the token, which are checked in order:

        1. Directly via this argument.
        2. If the argument starts with a dollar sign (``$``), it is
           treated as the name of an environment variable that points to
           a file containing the token. This is compatible with
           variables of type "File" in Gitlab CI/CD.
        3. A file called ``gitlab_artifacts_token`` in the
           ``$CREDENTIALS_DIRECTORY``.
        4. The ``CI_JOB_TOKEN`` environment variable. This is intended
           for use in Gitlab pipelines, in order to avoid having to
           create explicit tokens. Note that your instance might be set
           up with restrictive default permissions for the job token.
    project
        The path (e.g. ``my_username/my_cool_project``) or numeric ID of the
        project to pull the artifacts from. Defaults to the
        ``$CI_PROJECT_ID`` environment variable, which Gitlab sets to
        the project currently executing a pipeline.
    branch
        The branch to pull artifacts from. Defaults to the value of the
        ``CI_DEFAULT_BRANCH`` environment variable, or ``main`` if that
        is unset. Ignored if a numeric ID is given for ``job``.
    job
        Name of the job to pull artifacts from. May also be a numeric
        job ID.

        The last 20 pipeline runs on the given branch are searched for a
        successful job with this name that has artifacts. If nothing is
        found, an exception is raised, and a numeric job ID has to be
        specified explicitly.
    subdir
        An optional path prefix inside the artifacts archive to prepend
        to all file names.

    See Also
    --------
    capellambse.filehandler.http.HTTPFileHandler :
        A general-purpose HTTP file handler.
    """

    def __init__(
        self,
        path: str,
        *,
        subdir: str | pathlib.PurePosixPath = "/",
        token: str | None = None,
        project: str | int | None = None,
        branch: str | None = None,
        job: str | int,
    ) -> None:
        super().__init__(path, subdir=subdir)

        self.__path = self.__resolve_path(path)
        self.__token = self.__resolve_token(token)
        self.__project = self.__resolve_project(project)
        self.__branch = branch or os.getenv("CI_DEFAULT_BRANCH") or "main"
        self.__job = self.__resolve_job(job)

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}(path={self.__path!r}, token=<HIDDEN>,"
            f" project={self.__project!r}, branch={self.__branch!r},"
            f" job={self.__job!r}, subdir={self.subdir!r})"
        )

    @staticmethod
    def __resolve_path(path: str) -> str:
        if path == "glart:":
            if host := os.getenv("CI_SERVER_URL"):
                LOGGER.debug("Using current Gitlab instance: %s", host)
                return host
            else:
                LOGGER.debug("Using public Gitlab instance at gitlab.com")
                return "https://gitlab.com"

        if path.startswith("glart:"):
            return "https:" + path[6:]

        return path

    @classmethod
    def __resolve_token(cls, token: str | None) -> str:
        if token:
            return cls.__load_token_from_arg(token)

        if cred_dir := os.getenv("CREDENTIALS_DIRECTORY"):
            if token := cls.__load_token_from_credentials(cred_dir):
                return token

        if token := os.getenv("CI_JOB_TOKEN"):
            LOGGER.debug("Using the $CI_JOB_TOKEN")
            return token

        raise TypeError("Cannot connect to Gitlab: No 'token' found")

    @staticmethod
    def __load_token_from_arg(token: str) -> str:
        if not token.startswith("$"):
            return token

        try:
            filename = os.environ[token[1:]]
        except KeyError:
            raise ValueError(
                f"Token environment variable {token} is unset"
            ) from None

        try:
            contents = pathlib.Path(filename).read_text(encoding="ascii")
        except FileNotFoundError:
            raise ValueError(f"Token file from {token} not found") from None
        except OSError:
            raise ValueError(f"Cannot read token file from {token}") from None

        contents = contents.strip()
        if not contents:
            raise ValueError(f"Token file from {token} is empty")
        return contents

    @staticmethod
    def __load_token_from_credentials(cred_dir: str) -> str | None:
        cred_file = pathlib.Path(cred_dir, "gitlab_artifacts_token")
        try:
            token = cred_file.read_text(encoding="locale").strip()
        except OSError as err:
            LOGGER.debug("Cannot read token from %r: %s", cred_file, err)
        else:
            if token:
                LOGGER.debug("Using token from %s", cred_file)
                return token
            else:
                LOGGER.debug("Token file is empty: %s", cred_file)
        return None

    def __resolve_project(self, project: str | int | None) -> str | int:
        if isinstance(project, int):
            return project

        if project:
            project = urllib.parse.quote(project, safe="")
            info = self.__get(f"/projects/{project}")
            LOGGER.debug("Resolved project %r to ID %d", project, info["id"])
            return info["id"]

        if project := os.getenv("CI_PROJECT_ID"):
            return int(project)

        raise TypeError("No 'project' specified, and not running in Gitlab CI")

    def __resolve_job(self, job: str | int) -> int:
        if isinstance(job, int):
            return job

        for jobinfo in self.__iterget(
            f"/projects/{self.__project}/jobs?scope=success", max=200
        ):
            if (
                jobinfo["name"] == job
                and jobinfo["pipeline"]["ref"] == self.__branch
                and "artifacts_file" in jobinfo
            ):
                LOGGER.debug("Selected job with ID %d", jobinfo["id"])
                return jobinfo["id"]

        raise RuntimeError(f"No recent successful {job!r} job found")

    def __rawget(self, url: str) -> requests.Response:
        """Make a GET request and return the raw Response object.

        Parameters
        ----------
        url
            The full URL to request, including the server and protocol.
        """
        LOGGER.debug("GET %s", url)
        return requests.get(
            url, headers={"PRIVATE-TOKEN": self.__token}, timeout=30
        )

    def __get(self, path: str) -> t.Any:
        """Make a GET request and return the decoded JSON response."""
        response = self.__rawget(f"{self.__path}/api/v4{path}")
        LOGGER.debug("Status: %d", response.status_code)
        response.raise_for_status()
        return response.json()

    def __iterget(self, path: str, *, max: int = 0) -> cabc.Iterator[t.Any]:
        """Iterate over a result set, using multiple GET requests if necessary.

        This method implements the keyset-based pagination method.

        Parameters
        ----------
        path
            The base URL for the first request.
        max
            Maximum number of entries to yield. If 0, yield everything.
        """
        sep = ("?", "&")["?" in path]
        next_url = (
            f"{self.__path}/api/v4{path}{sep}pagination=keyset&per_page="
            + str(min(100, max or 100))
        )
        i = 0
        stop = max or math.inf

        while True:
            response = self.__rawget(next_url)
            response.raise_for_status()
            for i, object in enumerate(response.json(), start=i):
                yield object
                if i >= stop:
                    return

            match = RE_LINK_NEXT.fullmatch(response.headers["Link"])
            if not match:
                break
            next_url = match.group(0)

    def get_model_info(self) -> loader.ModelInfo:
        return loader.ModelInfo(branch=self.__branch, url=self.__path)

    def open(
        self,
        filename: str | pathlib.PurePosixPath,
        mode: t.Literal["r", "rb", "w", "wb"] = "rb",
    ) -> t.BinaryIO:
        path = str(helpers.normalize_pure_path(filename, base=self.subdir))

        if "w" in mode:
            raise TypeError("Cannot write to Gitlab artifacts")

        LOGGER.debug("Opening file %r for reading", path)
        response = self.__rawget(
            f"{self.__path}/api/v4/projects/{self.__project}"
            f"/jobs/{self.__job}/artifacts/{path}"
        )
        if response.status_code in (400, 404):
            raise FileNotFoundError(errno.ENOENT, filename)
        response.raise_for_status()
        return io.BytesIO(response.content)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0
"""Implements the Capella Filtering extension."""
from __future__ import annotations

import sys
import typing as t

import typing_extensions as te

import capellambse.model.common as c
from capellambse import model

VIEWPOINT: t.Final = "org.polarsys.capella.filtering"
NAMESPACE: t.Final = "http://www.polarsys.org/capella/filtering/6.0.0"
SYMBOLIC_NAME: t.Final = "filtering"

c.XTYPE_ANCHORS[__name__] = SYMBOLIC_NAME


@c.xtype_handler(None)
class FilteringCriterion(c.GenericElement):
    """A single filtering criterion."""

    _xmltag = "ownedFilteringCriteria"

    filtered_objects = c.ReferenceSearchingAccessor[c.GenericElement](
        (), "filtering_criteria", aslist=c.MixedElementList
    )


@c.xtype_handler(None)
class FilteringCriterionPkg(c.GenericElement):
    """A package containing multiple filtering criteria."""

    _xmltag = "ownedFilteringCriterionPkgs"

    criteria = c.DirectProxyAccessor(FilteringCriterion, aslist=c.ElementList)
    packages: c.Accessor[FilteringCriterionPkg]


@c.xtype_handler(None)
class FilteringModel(c.GenericElement):
    """A filtering model containing criteria to filter by."""

    criteria = c.DirectProxyAccessor(FilteringCriterion, aslist=c.ElementList)
    criterion_packages = c.DirectProxyAccessor(
        FilteringCriterionPkg, aslist=c.ElementList
    )


class AssociatedCriteriaAccessor(
    c.accessors.PhysicalAccessor[FilteringCriterion]
):
    def __init__(self) -> None:
        super().__init__(
            FilteringCriterion, aslist=c.ElementList[FilteringCriterion]
        )

    @t.overload
    def __get__(self, obj: None, objtype: type[t.Any]) -> te.Self:
        ...

    @t.overload
    def __get__(
        self, obj: c.ModelObject, objtype: type[t.Any] | None = ...
    ) -> c.ElementList[c.T]:
        ...

    def __get__(
        self,
        obj: c.ModelObject | None,
        objtype: type[t.Any] | None = None,
    ) -> te.Self | c.ElementList[c.T]:
        del objtype
        if obj is None:  # pragma: no cover
            return self

        loader = obj._model._loader
        try:
            xt_critset = f"{SYMBOLIC_NAME}:AssociatedFilteringCriterionSet"
            critset = next(loader.iterchildren_xt(obj._element, xt_critset))
        except StopIteration:
            elems = []
        else:
            links = critset.get("filteringCriteria", "")
            elems = list(loader.follow_links(obj._element, links))

        rv = self._make_list(obj, elems)
        if obj._constructed:
            sys.audit("capellambse.read_attribute", obj, self.__name__, rv)
            sys.audit("capellambse.getattr", obj, self.__name__, rv)
        return rv


def init() -> None:
    c.set_accessor(
        model.MelodyModel,
        "filtering_model",
        c.DirectProxyAccessor(FilteringModel, rootelem=model.XT_SYSENG),
    )
    c.set_accessor(
        c.GenericElement, "filtering_criteria", AssociatedCriteriaAccessor()
    )


c.set_self_references(
    (FilteringCriterionPkg, "packages"),
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

# pylint: disable=redefined-outer-name
from __future__ import annotations

import base64
import pathlib
import re
import shutil
import sys
from importlib import metadata

import pytest
import requests_mock

import capellambse

# pylint: disable-next=relative-beyond-top-level, useless-suppression
from .conftest import TEST_MODEL, TEST_ROOT  # type: ignore[import]

TEST_MODEL_5_0 = TEST_ROOT / "5_0" / TEST_MODEL


@pytest.mark.parametrize(
    "path",
    [
        pytest.param(str(TEST_MODEL_5_0), id="From string"),
        pytest.param(TEST_MODEL_5_0, id="From path"),
    ],
)
def test_model_loading_via_LocalFileHandler(path: str | pathlib.Path):
    capellambse.MelodyModel(path)


@pytest.mark.parametrize("suffix", [".capella", ".melodymodel"])
def test_model_loading_with_invalid_entrypoint_fails(suffix: str):
    with pytest.raises(ValueError, match="(?i)invalid entrypoint"):
        capellambse.MelodyModel(TEST_MODEL_5_0.with_suffix(suffix))


def test_model_loading_via_GitFileHandler():
    path = "git+" + pathlib.Path.cwd().as_uri()
    capellambse.MelodyModel(
        path, entrypoint="tests/data/melodymodel/5_0/Melody Model Test.aird"
    )


def test_model_loading_from_badpath_raises_FileNotFoundError():
    badpath = TEST_ROOT / "Missing.aird"
    with pytest.raises(FileNotFoundError):
        capellambse.MelodyModel(badpath)


class FakeEntrypoint:
    def __init__(self, expected_name, expected_url):
        self._expected_name = expected_name
        self._expected_url = expected_url

    @property
    def name(self):
        class AlwaysEqual:
            def __eq__(_, name):  # pylint: disable=E,I
                nonlocal self
                assert name == self._expected_name
                return True

        return AlwaysEqual()

    def load(self):
        def filehandler(url):
            assert url == self._expected_url

        return filehandler

    if sys.version_info < (3, 10):

        @classmethod
        def patch(cls, monkeypatch, expected_name, expected_url):
            eps = {
                "capellambse.filehandler": (cls(expected_name, expected_url),)
            }
            monkeypatch.setattr(metadata, "entry_points", lambda: eps)

    else:

        @classmethod
        def patch(cls, monkeypatch, expected_name, expected_url):
            def entry_points(group, name):
                assert group == "capellambse.filehandler"
                assert name == expected_name
                return (cls(expected_name, expected_url),)

            monkeypatch.setattr(metadata, "entry_points", entry_points)


def test_a_single_protocol_is_not_swallowed_by_get_filehandler(
    monkeypatch,
):
    FakeEntrypoint.patch(monkeypatch, "testproto", "testproto://url")

    capellambse.get_filehandler("testproto://url")


def test_a_wrapping_protocol_separated_by_plus_is_stripped(monkeypatch):
    FakeEntrypoint.patch(monkeypatch, "realproto", "wrappedproto://url")

    capellambse.get_filehandler("realproto+wrappedproto://url")


@pytest.mark.parametrize(
    "path",
    [
        "/data/model/model.aird",
        r"S:\model\model.aird",
        r"S:/model/model.aird",
        r"\\?\S:\model\model.aird",
        r"\\modelserver\model\model.aird",
    ],
)
def test_plain_file_paths_are_recognized_as_file_protocol(monkeypatch, path):
    FakeEntrypoint.patch(monkeypatch, "file", path)

    capellambse.get_filehandler(path)


@pytest.mark.parametrize(
    "url",
    [
        "myhost:myrepo.git",
        "git@host:",
        "git@host:repo.git",
        "git@host:path/to/repo",
        "git@host:/path/to/repo",
    ],
)
def test_the_scp_short_form_is_recognized_as_git_protocol(monkeypatch, url):
    FakeEntrypoint.patch(monkeypatch, "git", url)

    capellambse.get_filehandler(url)


@pytest.mark.parametrize(
    "link",
    [
        "xtype MelodyModelTest.aird#078b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "MelodyModelTest.aird#071b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "#071b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "xtype MelodyModel%20Test.aird#078b2c69-4352-4cf9-9ea5-6573b75e5eec",
        "MelodyModel%20Test.aird#071b2c69-4352-4cf9-9ea5-6573b75e5eec",
    ],
)
def test_MelodyLoader_follow_link_finds_target(link: str):
    loader = capellambse.loader.MelodyLoader(TEST_MODEL_5_0)

    with pytest.raises(KeyError):
        assert loader.follow_link(None, link) is not None


@pytest.mark.parametrize(
    ["path", "subdir", "req_url"],
    [
        (
            "https://example.com/~user",
            "/",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/~user/%s",
            "/",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/?file=%q",
            "/",
            "https://example.com/?file=demo%2Fmy%20model.aird",
        ),
        (
            "https://example.com/",
            "~user",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/%s",
            "~user",
            "https://example.com/~user/demo/my%20model.aird",
        ),
        (
            "https://example.com/?file=%q",
            "~user",
            "https://example.com/?file=~user%2Fdemo%2Fmy%20model.aird",
        ),
    ],
)
def test_http_file_handler_replaces_percent_s_percent_q(
    requests_mock: requests_mock.Mocker, path: str, subdir: str, req_url: str
) -> None:
    endpoint = requests_mock.get(req_url)

    file_handler = capellambse.get_filehandler(path, subdir=subdir)
    file_handler.open("demo/my model.aird", "rb").close()

    assert endpoint.called_once


def test_http_file_handler_hands_auth_to_server(
    requests_mock: requests_mock.Mocker,
) -> None:
    username = "testuser"
    password = "testpassword"
    auth_header = base64.standard_b64encode(
        f"{username}:{password}".encode("utf-8")
    ).decode("ascii")
    expected_headers = {"Authorization": f"Basic {auth_header}"}
    endpoint = requests_mock.get(
        "https://example.com/test.svg",
        request_headers=expected_headers,
    )

    file_handler = capellambse.get_filehandler(
        "https://example.com", username=username, password=password
    )
    file_handler.open("test.svg", "rb").close()

    assert endpoint.called_once


def test_http_file_handlers_passed_through_custom_headers(
    requests_mock: requests_mock.Mocker,
) -> None:
    expected_headers = {"X-Test-Header": "PASSED"}
    endpoint = requests_mock.get(
        "https://example.com/test.svg",
        request_headers=expected_headers,
    )

    file_handler = capellambse.get_filehandler(
        "https://example.com", headers=expected_headers
    )
    file_handler.open("test.svg", "rb").close()

    assert endpoint.called_once


@pytest.fixture
def model_path_with_patched_version(
    request: pytest.FixtureRequest, tmp_path: pathlib.Path
) -> pathlib.Path:
    """Indirect parametrized fixture for version patched model.

    Parameters
    ----------
    request
        Special pytest fixture for access to test context, exposing the
        ``param`` attribute on indirect parametrization.
    tmp_path
        Pytest fixture giving a temporary path.

    Returns
    -------
    aird_path
        Path to Capella ``.aird`` file.
    """
    tmp_dest = tmp_path / "model"
    ignored = shutil.ignore_patterns("*.license")
    shutil.copytree(TEST_MODEL_5_0.parent, tmp_dest, ignore=ignored)
    request_param: str | tuple[str, str] = request.param
    for suffix in (".aird", ".capella"):
        model_file = (tmp_dest / TEST_MODEL).with_suffix(suffix)
        if isinstance(request_param, tuple):
            re_params = request_param
        else:
            re_params = (r"5\.0\.0", request_param)

        patched = re.sub(*re_params, model_file.read_text(encoding="utf-8"))
        model_file.write_text(patched, encoding="utf-8")
    return model_file.with_suffix(".aird")


@pytest.mark.parametrize(
    "model_path_with_patched_version",
    ["1.3.0", "1.4.2", "6.1.0"],
    indirect=True,
)
def test_loading_model_with_unsupported_version_fails(
    model_path_with_patched_version: pathlib.Path,
) -> None:
    with pytest.raises(capellambse.UnsupportedPluginVersionError):
        capellambse.MelodyModel(model_path_with_patched_version)


@pytest.mark.parametrize(
    "model_path_with_patched_version",
    [("http://www.eclipse.org/sirius/1.1.0", "Unknown")],
    indirect=True,
)
def test_loading_model_with_unsupported_plugin_fails(
    model_path_with_patched_version: pathlib.Path,
) -> None:
    with pytest.raises(capellambse.UnsupportedPluginError):
        capellambse.MelodyModel(model_path_with_patched_version)


def test_model_info_contains_viewpoints_and_capella_version() -> None:
    loader = capellambse.loader.MelodyLoader(TEST_MODEL_5_0)

    info = loader.get_model_info()

    assert hasattr(info, "viewpoints")
    assert info.viewpoints["org.polarsys.capella.core.viewpoint"] == "5.0.0"
    assert info.capella_version == "5.0.0"



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Objects and relations for information capture and data modelling.

Information objects inheritance tree (taxonomy):

.. diagram:: [CDB] Information [Taxonomy]

Information object-relations map (ontology):

.. diagram:: [CDB] Information [Ontology]
"""

from __future__ import annotations

from ... import common as c
from ... import modeltypes
from .. import capellacommon, capellacore, modellingcore
from . import datatype, datavalue


@c.xtype_handler(None)
class Unit(c.GenericElement):
    """Unit."""

    _xmltag = "ownedUnits"


@c.xtype_handler(None)
class Association(c.GenericElement):
    """An Association."""

    _xmltag = "ownedAssociations"

    members: c.Accessor[Property]
    navigable_members: c.Accessor[Property]
    source_role: c.Accessor[Property]

    @property
    def roles(self) -> c.ElementList[Property]:
        roles = []
        if self.source_role:
            assert isinstance(self.source_role, c.GenericElement)
            roles.append(self.source_role._element)

        assert isinstance(self.navigable_members, c.ElementList)
        roles.extend(prop._element for prop in self.navigable_members)
        return c.ElementList(self._model, roles, Property)


@c.xtype_handler(None)
class PortAllocation(modellingcore.TraceableElement):
    """An exchange between a ComponentPort and FunctionalPort."""

    _xmltag = "ownedPortAllocations"


@c.xtype_handler(None)
class Property(c.GenericElement):
    """A Property of a Class."""

    _xmltag = "ownedFeatures"

    is_ordered = c.BooleanAttributeProperty(
        "ordered", __doc__="Indicates if property is ordered"
    )
    is_unique = c.BooleanAttributeProperty(
        "unique", __doc__="Indicates if property is unique"
    )
    is_abstract = c.BooleanAttributeProperty(
        "isAbstract", __doc__="Indicates if property is abstract"
    )
    is_static = c.BooleanAttributeProperty(
        "isStatic", __doc__="Indicates if property is static"
    )
    is_part_of_key = c.BooleanAttributeProperty(
        "isPartOfKey", __doc__="Indicates if property is part of key"
    )
    is_derived = c.BooleanAttributeProperty(
        "isDerived", __doc__="Indicates if property is abstract"
    )
    is_read_only = c.BooleanAttributeProperty(
        "isReadOnly", __doc__="Indicates if property is read-only"
    )
    visibility = c.EnumAttributeProperty(
        "visibility", modeltypes.VisibilityKind, default="UNSET"
    )
    kind = c.EnumAttributeProperty(
        "aggregationKind", modeltypes.AggregationKind, default="UNSET"
    )
    type = c.AttrProxyAccessor(c.GenericElement, "abstractType")
    default_value = c.RoleTagAccessor("ownedDefaultValue")
    min = c.RoleTagAccessor("ownedMinValue")
    max = c.RoleTagAccessor("ownedMaxValue")
    null_value = c.RoleTagAccessor("ownedNullValue")
    min_card = c.RoleTagAccessor("ownedMinCard")
    max_card = c.RoleTagAccessor("ownedMaxCard")
    association = c.ReferenceSearchingAccessor(Association, "roles")


@c.xtype_handler(None)
class Class(c.GenericElement):
    """A Class."""

    _xmltag = "ownedClasses"

    sub: c.Accessor
    super: c.Accessor[Class]
    is_abstract = c.BooleanAttributeProperty(
        "abstract", __doc__="Indicates if class is abstract"
    )
    is_final = c.BooleanAttributeProperty(
        "final", __doc__="Indicates if class is final"
    )
    is_primitive = c.BooleanAttributeProperty(
        "isPrimitive", __doc__="Indicates if class is primitive"
    )
    visibility = c.EnumAttributeProperty(
        "visibility", modeltypes.VisibilityKind, default="UNSET"
    )
    state_machines = c.DirectProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )
    owned_properties = c.DirectProxyAccessor(Property, aslist=c.ElementList)

    @property
    def properties(self) -> c.ElementList[Property]:
        """Return all owned and inherited properties."""
        return (
            self.owned_properties + self.super.properties
            if self.super is not None
            else self.owned_properties
        )


@c.xtype_handler(None)
class InformationRealization(modellingcore.TraceableElement):
    """A realization for a Class."""

    _xmltag = "ownedInformationRealizations"


@c.xtype_handler(None)
class Union(Class):
    """A Union."""

    _xmltag = "ownedClasses"

    kind = c.EnumAttributeProperty(
        "kind", modeltypes.UnionKind, default="UNION"
    )


@c.xtype_handler(None)
class Collection(c.GenericElement):
    """A Collection."""

    _xmltag = "ownedCollections"

    kind = c.EnumAttributeProperty(
        "kind", modeltypes.CollectionKind, default="ARRAY"
    )

    sub: c.Accessor
    super: c.Accessor[Collection]


@c.xtype_handler(None)
class DataPkg(c.GenericElement):
    """A data package that can hold classes."""

    classes = c.DirectProxyAccessor(Class, aslist=c.ElementList)
    unions = c.DirectProxyAccessor(Union, aslist=c.ElementList)
    collections = c.DirectProxyAccessor(Collection, aslist=c.ElementList)
    enumerations = c.DirectProxyAccessor(
        datatype.Enumeration, aslist=c.ElementList
    )
    complex_values = c.DirectProxyAccessor(
        datavalue.ComplexValue, aslist=c.ElementList
    )
    packages: c.Accessor


@c.xtype_handler(None)
class ExchangeItemElement(c.GenericElement):
    """An ExchangeItemElement (proxy link)."""

    _xmltag = "ownedElements"

    abstract_type = c.AttrProxyAccessor(c.GenericElement, "abstractType")
    owner: c.Accessor

    min_card = c.RoleTagAccessor("ownedMinCard")
    max_card = c.RoleTagAccessor("ownedMaxCard")


@c.xtype_handler(None)
class ExchangeItem(c.GenericElement):
    """An item that can be exchanged on an Exchange."""

    _xmltag = "ownedExchangeItems"

    type = c.EnumAttributeProperty(
        "exchangeMechanism", modeltypes.ExchangeItemType, default="UNSET"
    )
    elements = c.DirectProxyAccessor(ExchangeItemElement, aslist=c.ElementList)
    exchanges: c.Accessor[c.GenericElement]


for cls in [Class, Union, datatype.Enumeration, Collection]:
    c.set_accessor(
        cls,
        "super",
        c.LinkAccessor(
            "ownedGeneralizations", capellacore.Generalization, attr="super"
        ),
    )
    c.set_accessor(
        cls,
        "sub",
        c.ReferenceSearchingAccessor(cls, "super", aslist=c.MixedElementList),
    )

c.set_accessor(
    datavalue.EnumerationLiteral,
    "owner",
    c.ParentAccessor(datatype.Enumeration),
)
c.set_accessor(
    DataPkg, "packages", c.DirectProxyAccessor(DataPkg, aslist=c.ElementList)
)
c.set_accessor(ExchangeItemElement, "owner", c.ParentAccessor(ExchangeItem))
c.set_accessor(
    Association,
    "members",
    c.DirectProxyAccessor(Property, aslist=c.ElementList),
)
c.set_accessor(
    Association,
    "navigable_members",
    c.AttrProxyAccessor(Property, "navigableMembers", aslist=c.ElementList),
)
c.set_accessor(Association, "source_role", c.DirectProxyAccessor(Property))
c.set_accessor(
    Class,
    "realized_classes",
    c.LinkAccessor[Class](
        "ownedInformationRealizations",
        InformationRealization,
        aslist=c.ElementList,
        attr="targetElement",
    ),
)
c.set_accessor(
    Class,
    "realizations",
    c.DirectProxyAccessor(InformationRealization, aslist=c.ElementList),
)
c.set_accessor(
    Class,
    "realized_by",
    c.ReferenceSearchingAccessor(
        Class, "realized_classes", aslist=c.ElementList
    ),
)



# SPDX-FileCopyrightText: Copyright DB Netz AG and the capellambse contributors
# SPDX-License-Identifier: Apache-2.0

"""Functions that handle element styling."""
from __future__ import annotations

import collections.abc as cabc
import struct
import typing as t

import lxml.etree

from capellambse import diagram

from ._common import LOGGER


def apply_style_overrides(
    diagram_class: str | None,
    element_class: str,
    ostyle: lxml.etree._Element,
) -> dict[str, str | diagram.RGB | list[str | diagram.RGB]]:
    """Apply style overrides defined in the AIRD to a semantic element.

    Parameters
    ----------
    diagram_class
        The target diagram's style class
    element_class
        The class of this element, in the form of:
        ``{Edge|Box}.{class}``.
    ostyle
        An ownedStyle element.
    """

    def _to_rgb(
        ostyle: lxml.etree._Element, attrib: str
    ) -> diagram.RGB | None:
        color = ostyle.get(attrib)
        if color is not None:
            color = diagram.RGB.fromcsv(color)
        return color

    if diagram_class is None:
        return {}

    styleoverrides: dict[str, diagram.CSSdef] = {}

    # Background color
    color = _to_rgb(ostyle, "color")
    bgcolor = _to_rgb(ostyle, "backgroundColor")
    fgcolor = _to_rgb(ostyle, "foregroundColor")
    if color:
        styleoverrides["fill"] = color
    elif bgcolor or fgcolor:
        bgcolor = bgcolor or diagram.RGB(255, 255, 255)
        fgcolor = fgcolor or diagram.RGB(209, 209, 209)

        if bgcolor == fgcolor:
            styleoverrides["fill"] = bgcolor
        else:
            styleoverrides["fill"] = [bgcolor, fgcolor]

    # Foreground / font color
    styleoverrides["text_fill"] = _to_rgb(ostyle, "labelColor")

    linestyle = ostyle.attrib.get("lineStyle")
    if linestyle == "dash":
        styleoverrides["stroke-dasharray"] = "5"
    elif linestyle == "dot":
        styleoverrides["stroke-dasharray"] = "1, 3"
    elif linestyle is not None:
        LOGGER.warning("Ignoring unknown line style %s", linestyle)

    if element_class.startswith("Edge."):
        styleoverrides["stroke"] = _to_rgb(ostyle, "strokeColor")
        styleoverrides["stroke-width"] = ostyle.get("size")
    elif element_class.startswith("Box."):
        styleoverrides["stroke"] = _to_rgb(ostyle, "borderColor")
        styleoverrides["stroke-width"] = ostyle.get("borderSize")
    return _filter_default_styles(diagram_class, element_class, styleoverrides)


def apply_visualelement_styles(
    diagram_class: str, element_class: str, data_element: lxml.etree._Element
) -> dict[str, str | diagram.RGB | list[str | diagram.RGB]]:
    """Apply style overrides defined in the AIRD to a visual element.

    Parameters
    ----------
    diagram_class
        The target diagram's style class
    element_class
        The class of this element, in the form of:
        ``{Edge|Box}.{class}``
    data_element
        The ``<data>`` subtree's child element
    """
    styleoverrides: dict[str, t.Any] = {}

    def unpack_rgb(color: str, default: int) -> diagram.RGB:
        color_int = int(data_element.get(color, default))
        return diagram.RGB(
            *struct.unpack_from("3Bx", struct.pack("<i", color_int))
        )

    styleoverrides["stroke"] = unpack_rgb("lineColor", 0xFFFFFF)
    styleoverrides["fill"] = unpack_rgb("fillColor", 0xFFFFFF)
    styleoverrides["text_fill"] = unpack_rgb("fontColor", 0x000000)

    try:
        styleoverrides["stroke-width"] = int(data_element.attrib["lineWidth"])
    except KeyError:
        pass

    return _filter_default_styles(diagram_class, element_class, styleoverrides)


def _filter_default_styles(
    diagram_class: str,
    element_class: str,
    styleoverrides: cabc.Mapping[str, t.Any],
) -> dict[str, str | diagram.RGB | list[str | diagram.RGB]]:
    def style_is_default(key: str, val: t.Any) -> bool:
        for dia_lookup, elm_lookup in [
            (diagram_class, element_class),
            ("__GLOBAL__", element_class),
            (diagram_class, element_class.split(".")[0]),
            ("__GLOBAL__", element_class.split(".")[0]),
        ]:
            defstyle = diagram.STYLES.get(dia_lookup, {}).get(elm_lookup, {})
            try:
                return defstyle[key] == val
            except KeyError:
                pass
        return False

    return {
        k: [str(vx) for vx in v] if isinstance(v, list) else str(v)
        for k, v in styleoverrides.items()
        if v and not style_is_default(k, v)
    }
